from collections.abc import Iterator
from dataclasses import asdict, dataclass

import arxiv

from scidock.config import logger
from scidock.parsers.query_parser import clear_query, extract_arxiv_ids, extract_names
from scidock.search_engines.metadata import Metadata
from scidock.utils import dump_json, get_default_repository_path, load_json

client = arxiv.Client()

__all__ = ('search', 'download', 'extract_arxiv_ids')


@dataclass
class ArXivItem:
    title: str
    arxiv_id: str
    relevance_score: float = 1000.0

    def __post_init__(self):
        self.DOI = f'10.48550/arXiv.{self.arxiv_id}'

    def __str__(self):
        return f'{self.title.rstrip('.')}. DOI: {self.DOI}'


def search(query: str, extended: bool = False) -> Iterator[ArXivItem]:
    arxiv_ids = extract_arxiv_ids(query)
    logger.info(f'Extracted arXiv IDs: {arxiv_ids!r}')

    if arxiv_ids:
        search_request = arxiv.Search(id_list=arxiv_ids)
        for paper in client.results(search_request):
            yield ArXivItem(paper.title, paper.get_short_id())

        return

    search_query = ''

    names = extract_names(query)
    logger.info(f'Extracted names: {names!r}')
    if names is not None:
        search_query += 'au:' + ' AND '.join(names)

    # TODO: do something clever with extracting titles

    search_query += ('all:' if extended else 'ti:') + clear_query(query)
    search_request = arxiv.Search(query=search_query, sort_by=arxiv.SortCriterion.Relevance)

    for paper in client.results(search_request):
        yield ArXivItem(paper.title, paper.get_short_id())


def download(arxiv_id: str):
    search_request = arxiv.Search(id_list=[arxiv_id])
    paper = next(client.results(search_request))

    repository_path = get_default_repository_path()
    # noinspection PyProtectedMember
    filename = paper._get_default_filename()
    logger.info(f'Attempting to download a file with {filename = } for {arxiv_id = }')

    paper.download_pdf(dirpath=repository_path)

    content_path = f'{repository_path}/.scidock/content.json'
    repository_content = load_json(content_path)
    repository_content['local'][filename] = asdict(Metadata(paper.title, f'10.48550/arXiv.{paper.get_short_id()}'))
    dump_json(repository_content, content_path)



from collections.abc import Iterable, Iterator, Sequence
from typing import Any

import questionary
from questionary.prompts.common import Choice, InquirerControl
from rich.status import Status

__all__ = ('progress_bar',)

type ChoiceSequence = Sequence[str | Choice | dict[str, Any]]


class IterativeInquirerControl(InquirerControl):
    def __init__(self, choices: ChoiceSequence | Iterator | tuple[Iterable, Iterator], *args, **kwargs):
        choice_prefix = []

        if len(choices) == 2 and isinstance(choices[0], Iterable) and isinstance(choices[1], Iterator):  # noqa: PLR2004
            choice_prefix, choices = choices

        self.WINDOW_SIZE = (len(choice_prefix) // 10 + 1) * 10

        if isinstance(choices, Iterator):
            initial_choices = choice_prefix + [str(next(choices, '')) for _ in range(self.WINDOW_SIZE - len(choice_prefix))]
            initial_choices = list(filter(lambda choice: choice, initial_choices))

            if not initial_choices:
                raise ValueError('No choices provided')

            self.continuation_stream = choices
        else:
            initial_choices = choices
            self.continuation_stream = None

        self.trace = initial_choices
        self.trace_index = 0

        super().__init__(initial_choices, *args[1:], **kwargs)

    def select_previous(self) -> None:
        if self.continuation_stream is not None and self.pointed_at == 0:
            if self.trace_index == 0:
                return

            new_window = self.trace[self.trace_index - 1:self.trace_index + self.WINDOW_SIZE - 1]
            self.choices = [Choice.build(c) for c in new_window]
        else:
            self.pointed_at = (self.pointed_at - 1) % self.choice_count

        self.trace_index -= 1

    def select_next(self) -> None:
        if self.continuation_stream is not None:
            if self.trace_index == len(self.trace) - self.WINDOW_SIZE:
                self.trace.append(str(next(self.continuation_stream)))

            if self.pointed_at == self.choice_count - 1:
                new_window = self.trace[self.trace_index + 1:self.trace_index + self.WINDOW_SIZE + 1]
                self.choices = [Choice.build(c) for c in new_window]

        if self.choice_count == 0:
            self.choices = [Choice.build(self.trace[0])]
            self.pointed_at = 0
            self.trace_index = 0
            return

        self.pointed_at = (self.pointed_at + 1) % self.choice_count
        self.trace_index += 1


class ProgressBar(Status):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.previous_status = self.status

    def update(self, *args, **kwargs):
        self.previous_status = self.status
        super().update(*args, **kwargs)

    def revert_status(self):
        self.update(self.previous_status)


questionary.prompts.select.InquirerControl = IterativeInquirerControl

progress_bar = ProgressBar('Initializing...', spinner='dots2')



import json
import re

import requests
from bs4 import BeautifulSoup

from scidock.config import logger
from scidock.utils import filename_from_metadata, save_file_to_repo

__all__ = ('attempt_download',)


def attempt_download(doi: str, proxies: dict[str, str] | None = None) -> tuple[bool, str]:
    if proxies is None:
        proxies = {}
    headers = {
        'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/125.0.0.0 Safari/537.3'}
    logger.info(f'Attempting to follow a DOI redirect with DOI = {doi} and proxy configuration: {proxies}')

    publisher_page = requests.get(f'https://doi.org/{doi}', proxies=proxies, headers=headers, timeout=5)
    soup = BeautifulSoup(publisher_page.text, 'html.parser')

    if publisher_page.headers.get('Content-Type') in ('application/pdf', 'application/octet-stream'):
        logger.info('DOI redirected to a page with plain PDF')
        title = soup.title.text
        filename = filename_from_metadata(doi, title)

        return save_file_to_repo(publisher_page.url, filename, doi, title, 'DOI redirect', proxies), publisher_page.url

    if doi.startswith('10.1109') or 'IEEE' in soup.title.text:  # IEEE publisher
        logger.info('DOI redirected to a page of a known publisher: IEEE')

        metadata_pattern = re.compile(r'xplGlobal.document.metadata=(.*?});', re.MULTILINE | re.DOTALL)
        script = soup.find('script', string=metadata_pattern)
        metadata = json.loads(metadata_pattern.search(script.string).group(1))

        download_link = 'https://ieeexplore.ieee.org' + metadata['pdfPath']
        download_link = download_link.replace('iel', 'ielx')  # might be unstable
        title = metadata['displayDocTitle']
        filename = filename_from_metadata(doi, title)

        return save_file_to_repo(download_link, filename, doi, title, 'IEEE', proxies), download_link

    if doi.startswith('10.5772') or 'intechopen' in soup.title.text:  # IntechOpen publisher
        logger.info('DOI redirected to a page of a known publisher: IntechOpen')

        chapter_id = re.search(r'/chapters/(\d+)', publisher_page.url).group(1)
        download_link = 'https://www.intechopen.com' + '/chapter/pdf-download/' + chapter_id
        title = soup.find('h1', class_='title').text
        filename = filename_from_metadata(doi, title)

        return save_file_to_repo(download_link, filename, doi, title, 'IntechOpen', proxies), download_link

    if doi.startswith('10.3390') or 'mdpi' in soup.title.text:  # MDPI publisher
        logger.info('DOI redirected to a page of a known publisher: MDPI')

        download_link = publisher_page.url.rstrip('/') + '/pdf'
        title = soup.find('h1', class_='title').text
        filename = filename_from_metadata(doi, title)

        return save_file_to_repo(download_link, filename, doi, title, 'MDPI', proxies), download_link

    download_text_match = soup.find(string=re.compile(r'\bdownload\b', re.IGNORECASE))
    pdf_text_match = soup.find(string=re.compile('PDF', re.IGNORECASE))
    if download_text_match is not None or pdf_text_match is not None:
        return False, publisher_page.url

    return False, ''



import requests
from bs4 import BeautifulSoup

from scidock.config import logger
from scidock.utils import filename_from_metadata, save_file_to_repo

# TODO: make mirrors dynamic or more configurable
SCIHUB_MIRRORS = ['https://sci-hub.ru', 'https://sci-hub.se', 'https://sci-hub.st']
SCIDB_MIRRORS = ['https://annas-archive.gs/scidb', 'https://annas-archive.se/scidb']


def parse_scihub(soup: BeautifulSoup, doi: str, mirror: str) -> tuple[str | None, str | None, str | None]:
    download_button = soup.find('button')
    if not download_button:
        logger.info('Did not find the download button')
        return None, None, None

    if download_button.get('onclick') is None:
        logger.error('Download button does not have an "onclick" attribute (Sci-Hub DOM structure has changed)!')
        return None, None, None

    redirect_code = download_button['onclick']
    redirect_location = redirect_code.removeprefix("location.href='").removesuffix("'")
    download_link = ('https:' if redirect_location.startswith('//') else mirror) + redirect_location

    # TODO: improve citation parsing
    citation_container = soup.find('div', id='citation')
    citation_italic = citation_container.findChild('i')
    citation_title = citation_italic.text if citation_italic is not None else citation_container.text

    filename = filename_from_metadata(doi, citation_title)

    return download_link, filename, citation_title


def parse_scidb(soup: BeautifulSoup, doi: str) -> tuple[str | None, str | None, str | None]:
    download_button = soup.find('a', text='Download')
    if not download_button:
        logger.info('Did not find the download button')
        return None, None, None

    download_link = download_button['href']
    title_container = soup.find('div', class_='font-bold')
    title = title_container.text.removesuffix('🔍').strip()
    filename = filename_from_metadata(doi, title)

    return download_link, filename, title


def download(doi: str, proxies: dict[str, str] | None = None) -> bool:
    if proxies is None:
        proxies = {}
    logger.info(f'Attempting to download a file with DOI = {doi} and proxy configuration: {proxies}')

    for mirror in SCIHUB_MIRRORS + SCIDB_MIRRORS:
        try:
            # TODO: choose a sensible timeout based on the Internet speed
            timeout = 5 if proxies else 2
            preview_page = requests.get(f'{mirror}/{doi}', proxies=proxies, timeout=timeout, allow_redirects=False)
            break
        except requests.exceptions.Timeout:
            logger.debug(f'Timeout for the {mirror} Sci-Hub mirror')
            continue
    else:
        print('Unfortunately, all of the Sci-Hub mirrors are unavailable at your location. Try using a proxy')
        return False

    if preview_page.status_code in (301, 302):
        return False

    soup = BeautifulSoup(preview_page.text, 'html.parser')

    if 'sci-hub' in mirror:
        download_link, filename, title = parse_scihub(soup, doi, mirror)
        if any(field is None for field in (download_link, filename, title)):
            return False
    else:
        download_link, filename, title = parse_scidb(soup, doi)
        if any(field is None for field in (download_link, filename, title)):
            return False

    return save_file_to_repo(download_link, filename, doi, title, 'Sci-Hub', proxies)



from collections.abc import Iterator
from concurrent.futures import ThreadPoolExecutor
from dataclasses import dataclass
from pprint import pformat

import crossref.restful
import requests
from crossref.restful import Etiquette, Works

from scidock.config import logger
from scidock.parsers.mathml_parser import parse_document
from scidock.parsers.query_parser import clear_query, extract_dois, extract_keywords, extract_names, simplify_query
from scidock.utils import responsive_cache

crossref.restful.requests = requests.Session()

__all__ = ('search',)

etiquette = Etiquette('SciDock', '0.1.0', 'https://github.com/kgleba/scidock', 'kgleba@yandex.ru')
engine = Works(etiquette=etiquette)


@dataclass
class CrossRefItem:
    title: str
    DOI: str
    relevance_score: float = 1000.0

    def __str__(self):
        return f'{self.title.rstrip('.')}. DOI: {self.DOI}'


@responsive_cache
def perform_query(*args, **kwargs) -> Works:
    return engine.query(*args, **kwargs)


@responsive_cache
def prepare_query_args(query: str) -> tuple[list[str], dict[str, str]]:
    search_params = {}

    names = extract_names(query)
    logger.info(f'Extracted names: {names!r}')
    if names is not None:
        search_params['author'] = ' '.join(names)

    query = clear_query(query)
    keywords = extract_keywords(query)
    logger.info(f'Extracted keywords: {keywords!r}')
    if not keywords:
        keywords = [query]

    return keywords, search_params


def extract_metadata(paper: dict | None) -> CrossRefItem:
    if paper is None:
        paper = {}

    title = ' / '.join(paper.get('title', ('UNTITLED',)))

    if 'xmlns' in title:
        title = parse_document(title)

    return CrossRefItem(title, paper.get('DOI'), paper.get('score', 1000.0))


def search(query: str) -> tuple[int, Iterator[CrossRefItem]]:
    plain_query = simplify_query(query)

    first_search_result = None
    search_iter = iter([])
    n_search_results = 0

    if plain_query.strip():
        keywords, search_params = prepare_query_args(query)
        search_query = perform_query(*keywords, **search_params)
        search_iter = iter(search_query)

        logger.info('Launching concurrent requests')
        with ThreadPoolExecutor() as pool:
            count_future = pool.submit(search_query.count)
            search_future = pool.submit(lambda iterator: next(iterator, None), search_iter)

            n_search_results = count_future.result()
            first_search_result = search_future.result()

    def retrieve_papers() -> Iterator[CrossRefItem]:
        dois = extract_dois(query)
        for doi in dois:
            paper = engine.doi(doi)
            yield extract_metadata(paper)

        if first_search_result is not None:
            yield extract_metadata(first_search_result)

        for paper in search_iter:
            if None in (paper.get('title'), paper.get('DOI'), paper.get('score')):
                logger.warning(f'Received the paper with an unusual metadata: {pformat(paper)}')

            yield extract_metadata(paper)

    return n_search_results, next(iter(retrieve_papers, None))



import re
from typing import Any

import requests

from scidock.ui import progress_bar
from scidock.utils import responsive_cache

__all__ = ('extract_dois', 'extract_arxiv_ids', 'extract_names', 'extract_keywords', 'simplify_query', 'clear_query')

NLP_SERVER = 'https://kgleba-scidock-nlp.hf.space'

# following CrossRef's recommendation: https://www.crossref.org/blog/dois-and-matching-regular-expressions
DOI_PATTERN = re.compile(r'10.\d{4,9}/[-._;()/:a-zA-Z0-9]+')

# source: https://info.arxiv.org/help/arxiv_identifier_for_services.html
ARXIV_PATTERN = re.compile(r'(\d{4}.\d{4,5}|[a-z\-]+(\.[A-Z]{2})?/\d{7})(v\d+)?')
STRICT_ARXIV_PATTERN = re.compile(fr'arXiv\.{ARXIV_PATTERN.pattern}')

remote_data = {}


def _retrieve_remote_data(query: str, operation: str) -> Any:
    # updates relevant info about the `query` itself and `clear_query(query)`
    if remote_data.get(query) is None:
        progress_bar.update('Parsing your query using AI...')

        response = requests.post(f'{NLP_SERVER}/complex_analysis', json={'query': query}, timeout=10)
        remote_data.update(response.json())

        progress_bar.revert_status()

    return remote_data[query].get(operation)


@responsive_cache
def extract_dois(query: str) -> list[str]:
    return re.findall(DOI_PATTERN, query)


@responsive_cache
def extract_arxiv_ids(query: str, strict: bool = False, allow_overlap: bool = False) -> list[str]:
    if not allow_overlap:
        dois = extract_dois(query)
        for doi in dois:
            query = re.sub(f' *{doi} *', ' ', query)

    pattern = ARXIV_PATTERN
    if strict:
        pattern = STRICT_ARXIV_PATTERN

    return [''.join(match) for match in re.findall(pattern, query)]


def extract_names(query: str) -> list[str] | None:
    return _retrieve_remote_data(query, 'extract_names')


def extract_keywords(query: str) -> list[str] | None:
    return _retrieve_remote_data(query, 'extract_keywords')


def simplify_query(query: str) -> str | None:
    return _retrieve_remote_data(clear_query(query), 'remove_stop_words')


@responsive_cache
def clear_query(query: str) -> str:
    dois = extract_dois(query)
    for doi in dois:
        query = re.sub(f' *{doi} *', ' ', query)

    arxiv_ids = extract_arxiv_ids(query)
    for arxiv_id in arxiv_ids:
        query = re.sub(f' *{arxiv_id} *', ' ', query)

    names = extract_names(query)
    if names is not None:
        for name in names:
            query = re.sub(f' *{name} *', ' ', query)

    return query



import atexit
import logging
import logging.config
from functools import cache
from pathlib import Path

import yaml

__all__ = ('logger',)

logger = logging.getLogger('scidock')
Path('~/.scidock/logs').expanduser().mkdir(parents=True, exist_ok=True)


@cache  # ensure that the function gets called only once
def setup_logging():
    with open(Path(__file__).parent / 'logging.yaml') as config_file:
        log_config = yaml.load(config_file, Loader=yaml.SafeLoader)

    log_config['handlers']['file']['filename'] = str(Path(log_config['handlers']['file']['filename']).expanduser().absolute())

    logging.config.dictConfig(log_config)
    queue_handler = logging.getHandlerByName('queue_handler')
    if queue_handler is not None:
        queue_handler.listener.start()
        atexit.register(queue_handler.listener.stop)


setup_logging()



import re
import xml.etree.ElementTree as ET  # noqa: N817 - naming convention for ElementTree
from collections.abc import Callable
from functools import partial
from itertools import zip_longest
from typing import Any

from defusedxml import ElementTree as DET  # noqa: N814 - naming convention for ElementTree

NODE_TAGS = ('mfenced', 'mfrac', 'msub', 'msubsup', 'msup', 'mroot')
LEAF_TAGS = ('mi', 'mn', 'mo', 'msqrt', 'ms')
IRRELEVANT_TAGS = (
    'maction', 'menclose', 'merror', 'mglyph', 'mlabeledtr', 'mmultiscripts', 'mover', 'mpadded', 'mphantom', 'mrow', 'style', 'mspace',
    'mtable', 'mtd', 'mtext', 'mtr', 'mth', 'munder', 'munderover', 'semantics')


def parse_document(document: str):
    document = re.sub(r'mml:(\w+)', '\1', document)

    for irrelevant_tag in IRRELEVANT_TAGS:
        document = re.sub(f'<{irrelevant_tag}[^>]*>', '', document)
        document = re.sub(f'</{irrelevant_tag}>', '', document)

    index_delta = 0
    for math_tag in re.finditer(r'<math[^>]*>(.+?)</math>', document, re.MULTILINE | re.DOTALL):
        new_value = parse_tag(math_tag.group(0))
        start, end = math_tag.span()

        document_chars = list(document)
        document_chars[start + index_delta:end + index_delta] = list(new_value)
        document = ''.join(document_chars)

        index_delta += len(new_value) - (end - start)

    return re.sub('</math>', '', document)


def extract_source(element: ET.Element) -> str:
    tag = re.sub('({[^}]*})?([^>]*)', r'\2', element.tag)
    text = element.text if not list(element) else ''.join(extract_source(child) for child in element)
    formatted_attrs = ' '.join(f'{key}="{value}"' for key, value in element.items())
    return f'<{tag} {formatted_attrs}>{text}</{tag}>'


def dynamic_join(separators: str, source: Callable[[int], Any], n_elements: int | None = None) -> str:
    if n_elements is None:
        n_elements = len(separators) + 1

    if len(separators) != n_elements - 1:
        separators = separators[:-1] + separators[-1] * (n_elements - len(separators))

    elements = [str(source(i)) for i in range(n_elements)]

    joined_elements = ''
    for element, separator in zip_longest(elements, separators, fillvalue=''):
        joined_elements += element + separator

    return joined_elements


def parse_tag(tag: str) -> str:
    tag = tag.strip()

    root = DET.fromstring(tag)

    if root.tag.endswith('math'):
        return ''.join(parse_tag(extract_source(child)) for child in root)

    if root.tag in LEAF_TAGS:
        match root.tag:
            case 'mi' | 'mn' | 'mo':
                return root.text
            case 'ms':
                return f'"{root.text}"'
            case 'msqrt':
                return '√' + root.text

    join_sources = partial(dynamic_join, source=lambda i: parse_tag(extract_source(root[i])))
    tag_separators = {'mroot': '√', 'msub': '_', 'msup': '^', 'msubsup': '^_', 'mfrac': '/'}

    if root.tag in NODE_TAGS:
        match root.tag:
            case 'mfenced':
                open_, close, separators = root.attrib.get('open', '('), root.attrib.get('close', ')'), root.attrib.get('separators', ',')
                return open_ + join_sources(separators, n_elements=len(list(root))) + close
            case _:
                return join_sources(tag_separators[root.tag])

    return ''



import json
import random
import re
import string
from dataclasses import asdict
from functools import cache, wraps
from ipaddress import IPv4Address, IPv6Address
from os import PathLike
from pathlib import Path
from typing import Any

import requests
import tldextract

from scidock.config import logger
from scidock.search_engines.metadata import Metadata

KB = 1024

random.seed(42)


def load_json(filename: str | PathLike) -> Any:
    try:
        with open(filename, encoding='utf-8') as file:
            data = json.load(file)
    except (json.decoder.JSONDecodeError, FileNotFoundError):
        data = {}
    return data


def dump_json(data: Any, filename: str | PathLike) -> None:
    with open(filename, 'w', encoding='utf-8') as file:
        json.dump(data, file, ensure_ascii=False)


def extract_domain(url: str) -> str:
    url_metadata = tldextract.extract(url)
    return '.'.join((url_metadata.domain, url_metadata.suffix))


def get_default_repository_path() -> str | None:
    scidock_root = Path('~/.scidock').expanduser()
    repositories = load_json(scidock_root / 'config.json')

    if repositories.get('default') is not None:
        return repositories['repositories'][repositories['default']]['path']

    return None


def is_repository_initialized() -> bool:
    scidock_root = Path('~/.scidock').expanduser()

    if not (scidock_root / 'config.json').exists():
        return False

    repositories = load_json(scidock_root / 'config.json')
    return repositories.get('default') is not None


def require_initialized_repository(func):
    @wraps(func)
    def init_wrapper(*args, **kwargs):
        if not is_repository_initialized():
            logger.error('At least one repository should be initialized! See `scidock init --help`')
            return None

        return func(*args, **kwargs)

    return init_wrapper


def remove_outdated_repos(repositories: dict) -> dict:
    up_to_date_repositories = {}

    for repository_name, repository in repositories.items():
        if Path(repository['path']).exists():
            up_to_date_repositories[repository_name] = repository

    return up_to_date_repositories


def save_file_to_repo(download_link: str, filename: str, doi: str, title: str, caller_id: str,
                      proxies: dict[str, str] | None = None) -> bool:
    if proxies is None:
        proxies = {}
    headers = {
        'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/125.0.0.0 Safari/537.3'}

    logger.info(f'Attempting to download a file from {caller_id} with {filename = } and {download_link = } for {doi = }')

    repository_path = get_default_repository_path()
    try:
        download_page = requests.get(download_link, proxies=proxies, stream=True, headers=headers, timeout=5)
    except requests.exceptions.ConnectTimeout:
        logger.info(f'Download failed as {extract_domain(download_link)} is not responding')
        return False

    if download_page.status_code != 200:  # noqa: PLR2004 - the meaning and purpose of (status code) 200 are obvious from the context
        logger.info(f'Download failed with {download_page.status_code = }')
        return False

    content_type = download_page.headers.get('Content-Type')
    if content_type not in ('application/pdf', 'application/octet-stream'):
        logger.info(f'Download failed as Content-Type of the page is "{content_type}"')
        return False

    filename = re.sub('_+', '_', filename)
    filename = filename.replace('\r', '').replace('\n', '')

    with open(f'{repository_path}/{filename}', 'wb') as paper_file:
        for chunk in download_page.iter_content(chunk_size=10 * KB):
            paper_file.write(chunk)

    content_path = f'{repository_path}/.scidock/content.json'
    repository_content = load_json(content_path)
    repository_content['local'][filename] = asdict(Metadata(title, doi))
    dump_json(repository_content, content_path)

    return True


def filename_from_metadata(doi: str, title: str) -> str:
    title = title.strip()
    remove_punctuation = str.maketrans('', '', string.punctuation)
    filename = title.translate(remove_punctuation).replace(' ', '_')
    return '.'.join((doi, filename, 'pdf')).replace('/', '.')


def random_chain(*iterables, weights: list[float] | None = None):
    iterators = [iter(it) for it in iterables]

    while iterators:
        iterator = random.choices(iterators, weights, k=1)[0]

        try:
            yield next(iterator)
        except StopIteration:
            iterator_index = iterators.index(iterator)

            iterators.pop(iterator_index)
            if weights is not None:
                weights.pop(iterator_index)


def responsive_cache(func):
    func = cache(func)

    def format_args(args) -> str:
        return ', '.join(map(repr, args))

    def format_kwargs(kwargs) -> str:
        return ', '.join(f'{k}={v!r}' for k, v in kwargs.items())

    @wraps(func)
    def notification_wrapper(*args, **kwargs):
        hits = func.cache_info().hits
        func_return = func(*args, **kwargs)
        func_args = ', '.join((format_args(args), format_kwargs(kwargs)))

        logger.debug(f'Cache for {func.__name__}({func_args}) {'hit' if func.cache_info().hits > hits else 'missed'}')

        return func_return

    return notification_wrapper


def format_requests_proxy(proxy_type: str, ip: IPv4Address | IPv6Address, port: int):
    connection_string = f'{proxy_type}://{ip}:{port}'
    return {'http': connection_string, 'https': connection_string}


def get_current_proxy_setting():
    scidock_root = Path('~/.scidock').expanduser()
    current_config = load_json(scidock_root / 'config.json')
    return format_requests_proxy(*current_config['proxy'].values())

# TODO: consider creating context managers for working with different config files



import platform
import re
import subprocess
from collections.abc import Iterator
from dataclasses import asdict
from ipaddress import IPv4Address, IPv6Address
from pathlib import Path
from pprint import pformat

import click
import questionary
from click_params import IP_ADDRESS
from rapidfuzz import fuzz, process
from rapidfuzz.utils import default_process

from scidock.config import logger
from scidock.parsers.web_parser import attempt_download
from scidock.search_engines import arxiv_engine as arxiv
from scidock.search_engines import crossref_engine as crossref
from scidock.search_engines import scihub_engine as scihub
from scidock.search_engines.metadata import Metadata
from scidock.ui import progress_bar
from scidock.utils import (
    dump_json,
    get_current_proxy_setting,
    get_default_repository_path,
    load_json,
    random_chain,
    remove_outdated_repos,
    require_initialized_repository,
)

FUZZY_MATCH_RATE = 75


def update_recent_searches(paper: str):
    split_location = re.search(r'\. DOI: ', paper)
    title, doi = paper[:split_location.start()], paper[split_location.end():]

    repository_path = get_default_repository_path()
    content_path = f'{repository_path}/.scidock/content.json'

    repository_content = load_json(content_path)
    repository_content['recent_searches'][title] = asdict(Metadata(title, doi))
    dump_json(repository_content, content_path)


def split_search_results(query: str, arxiv_results: Iterator, search_results: Iterator) -> tuple[list, Iterator]:
    # approach of defining the cutoff value for CrossRef relevance scores
    search_prefix = []
    prefix_score_ratios = []
    prefix_max = -1
    previous_score = 0

    arxiv_ids = arxiv.extract_arxiv_ids(query)

    if arxiv_ids:
        search_prefix += list(map(str, arxiv_results))
    arxiv_results = iter(arxiv_results)
    first_arxiv_result = str(next(arxiv_results, ''))

    for search_result in search_results:
        search_prefix.append(str(search_result))

        prefix_max = max(prefix_max, search_result.relevance_score)
        prefix_score_ratios.append((search_result.relevance_score - previous_score) / prefix_max)
        if len(search_prefix) >= 8:  # noqa: PLR2004 - arbitrary number, should be tweaked afterwards
            best_score_ratio = prefix_score_ratios.index(max(prefix_score_ratios[1:]))
            insert_point = best_score_ratio + 2
            search_prefix.insert(insert_point, questionary.Separator())
            search_prefix[insert_point:insert_point] = [first_arxiv_result] + [str(next(arxiv_results, '')) for _ in range(4)]
            search_prefix = list(filter(lambda result: result, search_prefix))

            break

    search_results = random_chain(search_results, arxiv_results, weights=[0.4, 0.6])

    return search_prefix, search_results


@click.group()
def main():
    pass


@click.group()
def config():
    pass


# TODO: create `scidock test proxy`

@config.command('proxy')
@click.argument('proxy_type', type=click.Choice(['http', 'socks5'], case_sensitive=False))
@click.argument('ip', type=IP_ADDRESS)
@click.argument('port', type=int)
def proxy_configuration(proxy_type: str, ip: IPv4Address | IPv6Address, port: int):
    scidock_root = Path('~/.scidock').expanduser()
    current_config = load_json(scidock_root / 'config.json')
    current_config['proxy'] = {'type': proxy_type, 'ip': str(ip), 'port': port}
    dump_json(current_config, scidock_root / 'config.json')

    click.echo('Successfully configured proxy!')


def init(repository_path: Path, name: str | None = None):
    scidock_root = Path('~/.scidock').expanduser()
    scidock_repo_root = repository_path / '.scidock'
    if Path(scidock_repo_root).exists():
        click.echo('Repository in this directory is already initialized!', err=True)
        return

    scidock_repo_root.mkdir(parents=True)
    scidock_root.mkdir(exist_ok=True)

    current_config = load_json(scidock_root / 'config.json')

    if current_config.get('repositories') is None:
        current_config['repositories'] = {}

    current_config['repositories'] = remove_outdated_repos(current_config['repositories'])

    if name is not None:
        new_repository_name = name
    else:
        parts_included = 1
        new_repository_name = repository_path.absolute().parts[-1]
        while new_repository_name in current_config['repositories']:
            parts_included += 1
            new_repository_name = '/'.join(repository_path.parts[-parts_included:])

    new_repository_repr = {new_repository_name: {'path': str(repository_path.absolute())}}
    current_config['repositories'].update(new_repository_repr)
    current_config['default'] = new_repository_name
    current_config['proxy'] = {}

    dump_json({'local': {}, 'recent_searches': {}}, scidock_repo_root / 'content.json')
    dump_json(current_config, scidock_root / 'config.json')

    logger.info(f'Initialized repository with the following setup: {pformat(current_config)}')
    click.echo('Successfully initialized the repository!')


def download(query: str, proxies: dict[str, str] | None) -> bool:
    logger.info(f'Received download request with {query = }')

    query_dois = crossref.extract_dois(query)
    if len(query_dois) != 1:
        raise click.BadParameter('Target DOI is either not specified or ambiguous')

    progress_bar.start()
    progress_bar.update('Searching for a downloadable copy of the chosen paper...')

    target_doi = query_dois[0]

    target_arxiv_ids = arxiv.extract_arxiv_ids(target_doi, allow_overlap=True, strict=True)
    if target_arxiv_ids:
        arxiv.download(target_arxiv_ids[0])
        progress_bar.stop()
        click.echo('Successfully downloaded the paper!')
        return True

    if scihub.download(target_doi, proxies):
        progress_bar.stop()
        click.echo('Successfully downloaded the paper!')
        return True

    attempt_success, recommended_url = attempt_download(target_doi, proxies)
    if attempt_success:
        progress_bar.stop()
        click.echo('Successfully downloaded the paper!')
        return True

    progress_bar.stop()
    click.echo('A downloadable version of this work could not be found automatically :(')

    if recommended_url:
        click.echo(f'However, you could try and download the paper from the publisher\'s website manually: {recommended_url}')

    return False


def search(query: str, proxy: bool, extended: bool):
    # Suggested Workflow
    # Users get suggestions based on the relevance score provided by CrossRef
    # They are also provided with the option to open a pager (like GNU less) and scroll through more data generated on the fly
    # If nothing is to their liking, we can proceed searching for preprints in arXiv or in Google Scholar
    # The final goal of the search process is to retrieve the DOI, then one can proceed to the download stage

    proxies = {}
    if proxy:
        proxies = get_current_proxy_setting()

    progress_bar.start()
    progress_bar.update('Searching the CrossRef database...')

    n_search_results, search_results = crossref.search(query)
    logger.info(f'Number of search results for {query = !r}: {n_search_results:,}')

    progress_bar.update('Searching through the arXiv preprints...')

    arxiv_results = arxiv.search(query, extended)
    search_prefix, search_results = split_search_results(query, arxiv_results, search_results)

    progress_bar.stop()

    try:
        # noinspection PyTypeChecker
        # signature changes at a runtime, see `ui.IterativeInquirerControl`
        desired_paper = questionary.select(message='Choose the suitable paper to add to your library',
                                           choices=(search_prefix, search_results),
                                           pointer='\u276f').ask()
    except ValueError as e:
        if str(e) == 'No choices provided':
            click.echo('Nothing found! :(')
        return

    if desired_paper is not None:
        download_status = download(desired_paper, proxies)

        if not download_status:
            update_recent_searches(desired_paper)


def open_pdf(query: str):
    repository_path = get_default_repository_path()
    repository_content = load_json(f'{repository_path}/.scidock/content.json')

    query_dois = crossref.extract_dois(query)
    query_arxiv_ids = arxiv.extract_arxiv_ids(query)
    query_ids = query_dois + query_arxiv_ids
    if len(query_ids) > 1:
        raise click.BadParameter('Specified too many IDs: impossible to open single paper')

    filenames = list(repository_content['local'].keys())
    titles = [entry['title'] for entry in repository_content['local'].values()]
    dois = [entry['DOI'] for entry in repository_content['local'].values()]

    # noinspection PyTypeChecker
    # authors of the `rapidfuzz` library incorrectly specified the signature of the function
    best_title_match = process.extractOne(query, titles, scorer=fuzz.WRatio, score_cutoff=FUZZY_MATCH_RATE,
                                          processor=default_process)

    best_id_match = None
    if query_ids:
        # noinspection PyTypeChecker
        best_id_match = process.extractOne(query_ids[0], dois, scorer=fuzz.WRatio, score_cutoff=FUZZY_MATCH_RATE,
                                           processor=default_process)

    if best_title_match is None and best_id_match is None:
        click.echo('Did not find any relevant papers :(')
        return

    best_match = best_id_match if best_id_match is not None else best_title_match

    best_match_filename = filenames[best_match[2]]
    best_match_path = f'{repository_path}/{best_match_filename}'
    if ' ' in best_match_path:
        best_match_path = f'"{best_match_path}"'

    logger.info(f'Best Match Relevance Score: {best_match[1]}')

    # TODO: verify PDF header (to exclude the possibility of arbitrary code execution)
    # TODO: implement resolving full binary paths

    match platform.system():
        case 'Windows':
            subprocess.run(['powershell', '-Command', 'Invoke-Item', best_match_path], check=False)  # noqa: S603, S607 - see TODOs above
        case 'Linux':
            subprocess.run(['xdg-open', best_match_path], check=False)  # noqa: S603, S607 - see TODOs above
        case 'Darwin':
            subprocess.run(['open', best_match_path], check=False)  # noqa: S603, S607 - see TODOs above
        case _:
            logger.error(f'Operating system "{platform.system()}" not recognized!')
            return

    click.echo('Successfully opened the file!')


@click.command('init')
@click.argument('repository_path', type=click.Path(file_okay=False, path_type=Path))
@click.option('--name', type=str, default=None, help='Name of the repository. Defaults to the name of the folder')
def init_command(repository_path: Path, name: str | None):
    init(repository_path, name)


@click.command('search')
@click.argument('query', type=str)
@click.option('--proxy', is_flag=True, default=False, help='Whether to use a proxy in subsequent download requests')
@click.option('--extended', is_flag=True, default=False,
              help='Whether to include abstract and other fields in the search. Defaults to False (search by title only)')
@require_initialized_repository
def search_command(query: str, proxy: bool, extended: bool):
    search(query, proxy, extended)


@click.command('download')
@click.argument('DOI', type=str)
@click.option('--proxy', is_flag=True, default=False, help='Whether to use a proxy in download requests')
@require_initialized_repository
def download_command(doi: str, proxy: bool):
    proxies = {}
    if proxy:
        proxies = get_current_proxy_setting()

    download(doi, proxies)


@click.command('open')
@click.argument('query', type=str)
@require_initialized_repository
def open_command(query: str):
    open_pdf(query)


main.add_command(init_command)
main.add_command(search_command)
main.add_command(download_command)
main.add_command(open_command)

main.add_command(config)

if __name__ == '__main__':
    main()
